package javax.xml.bind;

import java.util.logging.ConsoleHandler;
import java.security.PrivilegedAction;
import java.util.Iterator;
import java.io.InputStream;
import java.util.Properties;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.security.AccessController;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.lang.reflect.Method;
import java.util.Map;
import java.net.URL;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

class ContextFinder
{
    private static final Logger logger;
    private static final String PLATFORM_DEFAULT_FACTORY_CLASS = "com.sun.xml.internal.bind.v2.ContextFactory";
    
    ContextFinder() {
    }
    
    private static void handleInvocationTargetException(InvocationTargetException x) throws JAXBException {
        Throwable t = x.getTargetException();
        if (t != null) {
            if ((t instanceof JAXBException)) {
                throw (JAXBException)t;
            }
            if ((t instanceof RuntimeException)) {
                throw (RuntimeException)t;
            }
            if ((t instanceof Error)) {
                throw (Error)t;
            }
        }
    }
    
    private static JAXBException handleClassCastException(Class originalType, Class targetType) {
        URL targetTypeURL = which(targetType);
        return new JAXBException(Messages.format("JAXBContext.IllegalCast", getClassClassLoader(originalType).getResource("javax/xml/bind/JAXBContext.class"), targetTypeURL));
    }
    
    static JAXBContext newInstance(String contextPath, String className, ClassLoader classLoader, Map properties) throws JAXBException {
        try {
            Class<Object> spFactory = safeLoadClass(className, classLoader);
            return newInstance(contextPath, spFactory, classLoader, properties);
        }
        catch (ClassNotFoundException x) {
            throw new JAXBException(Messages.format("ContextFinder.ProviderNotFound", className), x);
        }
        catch (RuntimeException x2) {
            throw x2;
        }
        catch (Exception x3) {
            throw new JAXBException(Messages.format("ContextFinder.CouldNotInstantiate", className, x3), x3);
        }
    }
    
    static JAXBContext newInstance(String contextPath, Class spFactory, ClassLoader classLoader, Map properties) throws JAXBException {
        try {
            Object context = null;
            try {
                Method m = spFactory.getMethod("createContext", new Class[] { String.class, ClassLoader.class, Map.class });
                context = m.invoke(null, new Object[] { contextPath, classLoader, properties });
            }
            catch (NoSuchMethodException ex) {}
            if (context == null) {
                Method m = spFactory.getMethod("createContext", new Class[] { String.class, ClassLoader.class });
                context = m.invoke(null, new Object[] { contextPath, classLoader });
            }
            if (!(context instanceof JAXBContext)) {
                throw handleClassCastException(context.getClass(), JAXBContext.class);
            }
            return (JAXBContext)context;
        }
        catch (InvocationTargetException x) {
            handleInvocationTargetException(x);
            Throwable e = x;
            if (x.getTargetException() != null) {
                e = x.getTargetException();
            }
            throw new JAXBException(Messages.format("ContextFinder.CouldNotInstantiate", spFactory, e), e);
        }
        catch (RuntimeException x2) {
            throw x2;
        }
        catch (Exception x3) {
            throw new JAXBException(Messages.format("ContextFinder.CouldNotInstantiate", spFactory, x3), x3);
        }
    }
    
    static JAXBContext newInstance(Class[] classes, Map properties, String className) throws JAXBException {
        ClassLoader cl = getContextClassLoader();
        Class<Object> spi;
        try {
            spi = safeLoadClass(className, cl);
        }
        catch (ClassNotFoundException e) {
            throw new JAXBException(e);
        }
        if (ContextFinder.logger.isLoggable(Level.FINE)) {
            ContextFinder.logger.log(Level.FINE, "loaded {0} from {1}", new Object[] { className, which(spi) });
        }
        return newInstance(classes, properties, spi);
    }
    
    static JAXBContext newInstance(Class[] classes, Map properties, Class spFactory) throws JAXBException {
        Method m;
        try {
            m = spFactory.getMethod("createContext", new Class[] { Class[].class, Map.class });
        }
        catch (NoSuchMethodException e) {
            throw new JAXBException(e);
        }
        try {
            Object context = m.invoke(null, new Object[] { classes, properties });
            if (!(context instanceof JAXBContext)) {
                throw handleClassCastException(context.getClass(), JAXBContext.class);
            }
            return (JAXBContext)context;
        }
        catch (IllegalAccessException e2) {
            throw new JAXBException(e2);
        }
        catch (InvocationTargetException e3) {
            handleInvocationTargetException(e3);
            Throwable x = e3;
            if (e3.getTargetException() != null) {
                x = e3.getTargetException();
            }
            throw new JAXBException(x);
        }
    }
    
    static JAXBContext find(String factoryId, String contextPath, ClassLoader classLoader, Map properties) throws JAXBException {
        String jaxbContextFQCN = JAXBContext.class.getName();
        StringTokenizer packages = new StringTokenizer(contextPath, ":");
        if (!packages.hasMoreTokens()) {
            throw new JAXBException(Messages.format("ContextFinder.NoPackageInContextPath"));
        }
        ContextFinder.logger.fine("Searching jaxb.properties");
        while (packages.hasMoreTokens()) {
            String packageName = packages.nextToken(":").replace('.', '/');
            StringBuilder propFileName = new StringBuilder().append(packageName).append("/jaxb.properties");
            Properties props = loadJAXBProperties(classLoader, propFileName.toString());
            if (props != null) {
                if (props.containsKey(factoryId)) {
                    String factoryClassName = props.getProperty(factoryId);
                    return newInstance(contextPath, factoryClassName, classLoader, properties);
                }
                throw new JAXBException(Messages.format("ContextFinder.MissingProperty", packageName, factoryId));
            }
            else {
                continue;
            }
        }
        ContextFinder.logger.fine("Searching the system property");
        String factoryClassName = (String)AccessController.doPrivileged(new GetPropertyAction("javax.xml.bind.context.factory"));
        if (factoryClassName != null) {
            return newInstance(contextPath, factoryClassName, classLoader, properties);
        }
        factoryClassName = (String)AccessController.doPrivileged(new GetPropertyAction(jaxbContextFQCN));
        if (factoryClassName != null) {
            return newInstance(contextPath, factoryClassName, classLoader, properties);
        }
        Class<Object> jaxbContext = lookupJaxbContextUsingOsgiServiceLoader();
        if (jaxbContext != null) {
            ContextFinder.logger.fine("OSGi environment detected");
            return newInstance(contextPath, jaxbContext, classLoader, properties);
        }
        ContextFinder.logger.fine("Searching META-INF/services");
        BufferedReader r = null;
        try {
            StringBuilder resource = new StringBuilder().append("META-INF/services/").append(jaxbContextFQCN);
            InputStream resourceStream = classLoader.getResourceAsStream(resource.toString());
            if (resourceStream != null) {
                r = new BufferedReader(new InputStreamReader(resourceStream, "UTF-8"));
                factoryClassName = r.readLine();
                if (factoryClassName != null) {
                    factoryClassName = factoryClassName.trim();
                }
                r.close();
                return newInstance(contextPath, factoryClassName, classLoader, properties);
            }
            else {
                ContextFinder.logger.log(Level.FINE, "Unable to load:{0}", resource.toString());
            }
        }
        catch (UnsupportedEncodingException e) {
            throw new JAXBException(e);
        }
        catch (IOException e2) {
            throw new JAXBException(e2);
        }
        finally {
            try {
                if (r != null) {
                    r.close();
                }
            }
            catch (IOException ex) {
                Logger.getLogger(ContextFinder.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        ContextFinder.logger.fine("Trying to create the platform default provider");
        return newInstance(contextPath, "com.sun.xml.internal.bind.v2.ContextFactory", classLoader, properties);
    }
    
    static JAXBContext find(Class[] classes, Map properties) throws JAXBException {
        String jaxbContextFQCN = JAXBContext.class.getName();
        for (Class c : classes) {
            ClassLoader classLoader = getClassClassLoader(c);
            Package pkg = c.getPackage();
            if (pkg != null) {
                String packageName = pkg.getName().replace('.', '/');
                String resourceName = packageName + "/jaxb.properties";
                ContextFinder.logger.log(Level.FINE, "Trying to locate {0}", resourceName);
                Properties props = loadJAXBProperties(classLoader, resourceName);
                if (props == null) {
                    ContextFinder.logger.fine("  not found");
                }
                else {
                    ContextFinder.logger.fine("  found");
                    if (props.containsKey("javax.xml.bind.context.factory")) {
                        String factoryClassName = props.getProperty("javax.xml.bind.context.factory").trim();
                        return newInstance(classes, properties, factoryClassName);
                    }
                    throw new JAXBException(Messages.format("ContextFinder.MissingProperty", packageName, "javax.xml.bind.context.factory"));
                }
            }
        }
        ContextFinder.logger.log(Level.FINE, "Checking system property {0}", "javax.xml.bind.context.factory");
        String factoryClassName = (String)AccessController.doPrivileged(new GetPropertyAction("javax.xml.bind.context.factory"));
        if (factoryClassName != null) {
            ContextFinder.logger.log(Level.FINE, "  found {0}", factoryClassName);
            return newInstance(classes, properties, factoryClassName);
        }
        ContextFinder.logger.fine("  not found");
        ContextFinder.logger.log(Level.FINE, "Checking system property {0}", jaxbContextFQCN);
        factoryClassName = (String)AccessController.doPrivileged(new GetPropertyAction(jaxbContextFQCN));
        if (factoryClassName != null) {
            ContextFinder.logger.log(Level.FINE, "  found {0}", factoryClassName);
            return newInstance(classes, properties, factoryClassName);
        }
        ContextFinder.logger.fine("  not found");
        Class<Object> jaxbContext = lookupJaxbContextUsingOsgiServiceLoader();
        if (jaxbContext != null) {
            ContextFinder.logger.fine("OSGi environment detected");
            return newInstance(classes, properties, jaxbContext);
        }
        ContextFinder.logger.fine("Checking META-INF/services");
        BufferedReader r = null;
        try {
            String resource = "META-INF/services/" + jaxbContextFQCN;
            ClassLoader classLoader2 = getContextClassLoader();
            URL resourceURL;
            if (classLoader2 == null) {
                resourceURL = ClassLoader.getSystemResource(resource);
            }
            else {
                resourceURL = classLoader2.getResource(resource);
            }
            if (resourceURL != null) {
                ContextFinder.logger.log(Level.FINE, "Reading {0}", resourceURL);
                r = new BufferedReader(new InputStreamReader(resourceURL.openStream(), "UTF-8"));
                factoryClassName = r.readLine();
                if (factoryClassName != null) {
                    factoryClassName = factoryClassName.trim();
                }
                return newInstance(classes, properties, factoryClassName);
            }
            else {
                ContextFinder.logger.log(Level.FINE, "Unable to find: {0}", resource);
            }
        }
        catch (UnsupportedEncodingException e) {
            throw new JAXBException(e);
        }
        catch (IOException e2) {
            throw new JAXBException(e2);
        }
        finally {
            if (r != null) {
                try {
                    r.close();
                }
                catch (IOException ex) {
                    ContextFinder.logger.log(Level.FINE, "Unable to close stream", ex);
                }
            }
        }
        ContextFinder.logger.fine("Trying to create the platform default provider");
        return newInstance(classes, properties, "com.sun.xml.internal.bind.v2.ContextFactory");
    }
    
    private static Class lookupJaxbContextUsingOsgiServiceLoader() {
        try {
            Class<Object> target = Class.forName("com.sun.org.glassfish.hk2.osgiresourcelocator.ServiceLoader");
            Method m = target.getMethod("lookupProviderClasses", new Class[] { Class.class });
            Iterator<Object> iter = ((Iterable)m.invoke(null, new Object[] { JAXBContext.class })).iterator();
            return iter.hasNext() ? ((Class)iter.next()) : null;
        }
        catch (Exception e) {
            ContextFinder.logger.log(Level.FINE, "Unable to find from OSGi: javax.xml.bind.JAXBContext");
            return null;
        }
    }
    
    private static Properties loadJAXBProperties(ClassLoader classLoader, String propFileName) throws JAXBException {
        Properties props = null;
        try {
            URL url;
            if (classLoader == null) {
                url = ClassLoader.getSystemResource(propFileName);
            }
            else {
                url = classLoader.getResource(propFileName);
            }
            if (url != null) {
                ContextFinder.logger.log(Level.FINE, "loading props from {0}", url);
                props = new Properties();
                InputStream is = url.openStream();
                props.load(is);
                is.close();
            }
        }
        catch (IOException ioe) {
            ContextFinder.logger.log(Level.FINE, "Unable to load " + propFileName, ioe);
            throw new JAXBException(ioe.toString(), ioe);
        }
        return props;
    }
    
    static URL which(Class clazz, ClassLoader loader) {
        String classnameAsResource = clazz.getName().replace('.', '/') + ".class";
        if (loader == null) {
            loader = getSystemClassLoader();
        }
        return loader.getResource(classnameAsResource);
    }
    
    static URL which(Class clazz) {
        return which(clazz, getClassClassLoader(clazz));
    }
    
    private static Class safeLoadClass(String className, ClassLoader classLoader) throws ClassNotFoundException {
        ContextFinder.logger.log(Level.FINE, "Trying to load {0}", className);
        try {
            SecurityManager s = System.getSecurityManager();
            if (s != null) {
                int i = className.lastIndexOf(46);
                if (i != -1) {
                    s.checkPackageAccess(className.substring(0, i));
                }
            }
            if (classLoader == null) {
                return Class.forName(className);
            }
            return classLoader.loadClass(className);
        }
        catch (SecurityException se) {
            if ("com.sun.xml.internal.bind.v2.ContextFactory".equals(className)) {
                return Class.forName(className);
            }
            throw se;
        }
    }
    
    private static ClassLoader getContextClassLoader() {
        if (System.getSecurityManager() == null) {
            return Thread.currentThread().getContextClassLoader();
        }
        return (ClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
            @Override
            public Object run() {
                return Thread.currentThread().getContextClassLoader();
            }
        });
    }
    
    private static ClassLoader getClassClassLoader(final Class c) {
        if (System.getSecurityManager() == null) {
            return c.getClassLoader();
        }
        return (ClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
            @Override
            public Object run() {
                return c.getClassLoader();
            }
        });
    }
    
    private static ClassLoader getSystemClassLoader() {
        if (System.getSecurityManager() == null) {
            return ClassLoader.getSystemClassLoader();
        }
        return (ClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
            @Override
            public Object run() {
                return ClassLoader.getSystemClassLoader();
            }
        });
    }
    
    static {
        logger = Logger.getLogger("javax.xml.bind");
        try {
            if (AccessController.doPrivileged(new GetPropertyAction("jaxb.debug")) != null) {
                ContextFinder.logger.setUseParentHandlers(false);
                ContextFinder.logger.setLevel(Level.ALL);
                ConsoleHandler handler = new ConsoleHandler();
                handler.setLevel(Level.ALL);
                ContextFinder.logger.addHandler(handler);
            }
        }
        catch (Throwable t) {}
    }
}
